"""A package for encoding and decoding weather dataset."""

import asyncio
import copy
import io
import os
import pickle
import random
import threading
import time
from collections import deque
from functools import partial
from time import perf_counter
from typing import (
    Any,
    BinaryIO,
    Deque,
    Dict,
    Generator,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

import aiofiles
import numpy as np
import yaml

from seqrecord.utils import (
    CONSUMER_SLEEP_INTERVAL,
    PRODUCER_SLEEP_INTERVAL,
    FileManager,
    LRUCache,
    TimeTracker,
    WriterBuffer,
)

MAX_RECORDFILE_SIZE = 1e9  # 1e8, 100 mb, maximum size of a single record file

WSR = TypeVar("WSR", bound="WSeqRecord")

# todo: record file size and write buffer size, how to choose it, and tune it
# todo: test effect of caching


# todo: use azure uri and manage sources of storage account by ourselves (replica of dataset)
# todo: MISSING, some transform work, subsample etc. Need to make sure data is same as produced by existing dataset


def recordfileidx2path(recorddir: str, file_idx: int) -> str:
    return os.path.join(recorddir, f"record_{file_idx}.bin")


class _PrefetchData:
    def __init__(self, source_data_generator, buffer_size: int):
        self.run_prefetcher = True
        # python deque is thread safe for appends and pops from opposite sides.
        # ref: https://stackoverflow.com/questions/8554153/is-this-deque-thread-safe-in-python
        self.prefetch_buffer: Deque = deque()
        self.buffer_size: int = buffer_size
        self.source_data_generator = source_data_generator


class WSeqRecord:
    """A serialization protocal that stores a single continuous long sequence of weather data into record files, while provides metadata of each frame to enable efficient random access of frames."""

    def __init__(
        self,
        recorddir: str,
    ) -> None:
        """_summary_

        Args:
            recorddir (str): directory record files is placed
        """
        # folder that stores data in a separate directory (subfolder)
        self.recorddir: str = recorddir
        os.makedirs(self.recorddir, exist_ok=True)
        self.features_written = None
        self.num_bytes: int = 0  # number of bytes in current file
        # track the frame idx endpoints for each record file, [[start_idx, end_idx)], includes start, exclusive on end
        self.idx_range_of_files: List[Tuple[int, int]] = []
        # file object for current record file
        self.file_desc: Optional[BinaryIO] = None
        # serialization proto info of each data item
        self.metadata: Dict[int, dict] = {}
        # index of current data item to be processed
        self.frame_idx: int = 0

        # write buffer # todo: determine the size of write buffer
        self.write_buffer = WriterBuffer()
        self.timer = TimeTracker()

    # todo add append_item(), add new frames to existing dataset (probably saved already)
    # todo add preput worker
    def write_item(
        self,
        frame: Dict[str, np.ndarray],
    ) -> None:
        """write one frame data dict(feature->np.ndarray) into byte repr and write bytes into
        current file. It has to be called by users in a sequential order. Every key,value in frame will
        be written into file.

        #todo: add default settings, where features are all features in metadata[idx]

        Args:
            frame (Dict[str, np.ndarray]): feature to data (np.ndarray)
        """
        if self.features_written is None:
            self.features_written = [key for key in frame]
        file_idx = len(self.idx_range_of_files) - 1
        if self.num_bytes == 0:
            # new file
            self.idx_range_of_files.append([self.frame_idx])
            file_idx += 1
            self.file_desc = open(
                recordfileidx2path(self.recorddir, file_idx),
                mode="wb",
            )
            # todo: check here if we are appending, if it is, we need to open a new file

        # get file name and start position for frame
        self.metadata[self.frame_idx] = {
            "frame_idx": self.frame_idx,
            "file_idx": file_idx,
            "bytes_offset": self.num_bytes,
        }
        num_bytes_in_frame = 0
        for feature, data in frame.items():
            self.metadata[self.frame_idx][feature] = {
                "is_none": (
                    data.dtype == np.dtype("O") and data == None
                ),  # this feature is essentially missing, and
                "dtype": data.dtype,
                "shape": data.shape,
                "bytes_offset": num_bytes_in_frame,
                "nbytes": data.nbytes,
            }
            self.write_buffer.write(data.tobytes())
            num_bytes_in_frame += data.nbytes

        self.metadata[self.frame_idx]["nbytes"] = num_bytes_in_frame
        self.num_bytes += num_bytes_in_frame
        self.frame_idx += 1

        # if self.write_buffer.is_full():
        #     self.file_desc.write(self.write_buffer.getbuffer())
        #     self.write_buffer.clear()
        #     self.file_desc.flush()

        if self.num_bytes > MAX_RECORDFILE_SIZE:
            # current file is big enough
            self.num_bytes = 0
            self.idx_range_of_files[-1].append(self.frame_idx)
            self.file_desc.write(self.write_buffer.getbuffer())
            self.write_buffer.clear()
            self.file_desc.flush()
            self.file_desc.close()
        return

    def close_recordfile(self):
        """Close opened file descriptor!

        This needs to be called by user when finishes scanning over the dataset.
        """
        if len(self.idx_range_of_files[-1]) == 1:
            # the last record file has not reached maximum file size
            self.idx_range_of_files[-1].append(self.frame_idx)

            # write what's left in self.writer buffer
            self.file_desc.write(self.write_buffer.getbuffer())
            self.write_buffer.close()

        self.file_desc.close()
        self.write_buffer = None
        self.file_desc = None
        self.num_frames = self.frame_idx
        self.num_files = len(self.idx_range_of_files)

    def recordfile_generator(self, frame_generator: callable):
        try:
            write_buffer = WriterBuffer()
            num_bytes = 0
            self.idx_range_of_files = []
            frame_idx = 0
            file_idx = 0
            for frame in frame_generator:
                if self.features_written is None:
                    self.features_written = [key for key in frame]
                if num_bytes == 0:
                    # new file
                    self.idx_range_of_files.append([frame_idx])
                self.metadata[frame_idx] = {
                    "frame_idx": frame_idx,
                    "file_idx": file_idx,
                    "bytes_offset": num_bytes,
                }
                num_bytes_in_frame = 0
                for feature, data in frame.items():
                    self.metadata[frame_idx][feature] = {
                        "is_none": (
                            data.dtype == np.dtype("O") and data == None
                        ),  # this feature is essentially missing, and
                        "dtype": data.dtype,
                        "shape": data.shape,
                        "bytes_offset": num_bytes_in_frame,
                        "nbytes": data.nbytes,
                    }
                    write_buffer.write(data.tobytes())
                    num_bytes_in_frame += data.nbytes

                self.metadata[frame_idx]["nbytes"] = num_bytes_in_frame
                frame_idx += 1
                num_bytes += num_bytes_in_frame
                if num_bytes > MAX_RECORDFILE_SIZE:
                    # current file is big enough
                    num_bytes = 0
                    self.idx_range_of_files[-1].append(frame_idx)
                    write_buffer.clear()
                    yield (file_idx, write_buffer.getvalue())
                    file_idx += 1

            if len(self.idx_range_of_files[-1]) == 1:
                # there is content left in the write_buffer
                self.idx_range_of_files[-1].append(frame_idx)
                yield (file_idx, write_buffer.getvalue())
        finally:
            write_buffer.close()
            self.num_files = len(self.idx_range_of_files)
            self.num_frames = frame_idx

    def put_frame(self, frame_generator: callable, prefetch_buffer_size: int = 5):
        # should be only adding frames here
        # two threads this function keep writing and send them to buffer
        # a separate thread writes the buffer to files as long as the buffer is non-empty
        try:
            prefetch_data = _PrefetchData(
                self.recordfile_generator(frame_generator=frame_generator),
                prefetch_buffer_size,
            )
            thread = threading.Thread(
                target=WSeqRecord.prefetch_thread_worker,
                args=(prefetch_data,),
                daemon=True,
            )
            thread.start()
            while prefetch_data.run_prefetcher:
                if len(prefetch_data.prefetch_buffer) > 0:
                    file_idx, content = prefetch_data.prefetch_buffer.popleft()
                    with open(recordfileidx2path(self.recorddir, file_idx), "wb") as f:
                        f.write(content)
                else:
                    # TODO: Calculate sleep interval based on previous availability speed
                    time.sleep(CONSUMER_SLEEP_INTERVAL)
        finally:
            prefetch_data.run_prefetcher = False
            if thread is not None:
                thread.join()
                thread = None

    def read_frame(
        self,
        file_desc: Union[io.BufferedReader, BinaryIO],
        metadata_frame: Dict[str, Union[int, dict]],
        features: List[str],
    ) -> Dict[str, np.ndarray]:
        """Given record file descriptor and serialization proto of a single frame, return the
        decoded dictionary(feature->data(np.ndarray)) of the item.

        Args:
            file_desc (io.BufferedReader): python file object of the record file (required by numpy)
            metadata_frame (Dict[str, Any]): dict that contains meta info of a specific frame
            features (List[str]):  features requested for frame
        Returns:
            Dict[str, np.ndarray]: data
        """
        frame = {}
        frame_offset = metadata_frame["bytes_offset"]
        for feature in features:
            frame[feature] = np.memmap(
                file_desc,
                dtype=metadata_frame[feature]["dtype"],
                mode="r",
                offset=frame_offset + metadata_frame[feature]["bytes_offset"],
                shape=metadata_frame[feature]["shape"],
            )
        return frame

    def iterate_frames(
        self, features: List[str]
    ) -> Generator[Dict[str, np.ndarray], None, None]:
        """Iterate sequentially over frames in the dataset

        Args:
            features (List[str]): a list of feature names requested from frames

        Returns:
            _type_: _description_

        Yields:
            Generator[Dict[str, np.ndarray], None, None]: generates one-frame data
        """
        assert all(
            len(endpoints) == 2 for endpoints in self.idx_range_of_files
        ), "Record file endpoints are not complete!"

        for file_idx in range(self.num_files):
            file_desc = open(
                recordfileidx2path(self.recorddir, file_idx=file_idx), mode="rb"
            )
            endpoints = self.idx_range_of_files[file_idx]
            for idx in range(endpoints[0], endpoints[1]):
                frame = self.read_frame(file_desc, self.metadata[idx], features)

                yield {feature: frame[feature] for feature in features}
            file_desc.close()

    # todoL add prefetch using threads or asyncio
    # todo: test effect of caching on real data
    def iterate_frame_pairs(
        self,
        input_features: List[str],
        target_features: List[str],
        max_pred_steps: int,
        filedesc_cache_cap: int = 10,
        frame_cache_cap: int = 20,
    ) -> Generator[Dict[str, np.ndarray], None, None]:
        """Iterate frames over the whole dataset

        # todo: to think about, if we don't shuffle files, then cache based on frame idx is convenient and effective.
        Args:
            input_features [List[str]]: a list of features requested for input
            target_features [List[str]]: a list of features requested for target
            max_pred_steps [int]: maximum number of leap steps for predictive frame
        Yields:
            Generator[Dict[str, np.ndarray], None, None]: data item [feature->data]. All data items are being returned sequentially
        """
        file_manager = FileManager(
            fileidx2path=partial(recordfileidx2path, recorddir=self.recorddir),
            cache_capacity=filedesc_cache_cap,
        )
        # given that, input and target features do not overlap, we only cache target frame
        # LRU might not be suitable, evicting based on idx seems better
        frame_cache = LRUCache(frame_cache_cap)
        assert all(
            len(endpoints) == 2 for endpoints in self.idx_range_of_files
        ), "Record file endpoints are not complete!"
        num_frames = len(self.metadata)  # or self.num_frames
        for fileidx4input in range(self.num_files):
            filedesc4input = file_manager.open_file(file_idx=fileidx4input)
            endpoints = self.idx_range_of_files[fileidx4input]
            # no target frame to predict for the last frame
            for frameidx4input in range(
                endpoints[0],
                min(endpoints[1], num_frames - 1),  # self.num_frames
            ):
                input_frame = self.read_frame(
                    filedesc4input, self.metadata[frameidx4input], input_features
                )
                # get the target frame for prediction, both start, stop inclusive
                lookahead_steps = min(
                    random.randint(1, max_pred_steps),
                    num_frames - 1 - frameidx4input,
                )
                frameidx4target = frameidx4input + lookahead_steps
                target_frame = frame_cache.get(frameidx4target)
                if target_frame is None:
                    fileidx4target = self.metadata[frameidx4target]["file_idx"]
                    filedesc4target = file_manager.open_file(fileidx4target)
                    target_frame = self.read_frame(
                        filedesc4target,
                        self.metadata[frameidx4target],
                        target_features,
                    )
                # colllate input and target frames so that input and target frame are np.ndarray
                input_frame = np.vstack(
                    [input_frame[feature] for feature in input_features]
                )
                target_frame = np.vstack(
                    [target_frame[feature] for feature in target_features]
                )
                # print(self.timer.summarize())
                yield {
                    "input": input_frame,
                    "target": target_frame,
                    "lookahead_steps": np.asarray(lookahead_steps),
                    "input_features": input_features,
                    "target_features": target_features,
                }
        file_manager.close_all_files()

    def async_iterate_frame_pairs(
        self,
        input_features: List[str],
        target_features: List[str],
        max_pred_steps: int,
        filedesc_cache_cap: int = 10,
    ) -> Generator[Dict[str, np.ndarray], None, None]:
        """Asyncly read two frames from (possibly) two files.

        Notes:
            No frame cache

        Returns:
            _type_: _description_

        Yields:
            _type_: _description_
        """
        # setup a single event loop for async read
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        # file_desc_cache is only used for target frame, since we are iterating for the input frame,
        file_desc_cache = LRUCache(capacity=filedesc_cache_cap)
        # given that, input and target features do not overlap, we only cache target frame
        # LRU might not be suitable, evicting based on idx seems better
        assert all(
            len(endpoints) == 2 for endpoints in self.idx_range_of_files
        ), "Record file endpoints are not complete!"

        num_frames = len(self.metadata)
        # read two frames using asyn io
        # file cache should only be used for future frames, since base file desc is used continuously
        try:
            for fileidx4input in range(self.num_files):
                filedesc4input = None
                endpoints = self.idx_range_of_files[fileidx4input]

                # no target frame to predict for the last frame
                for frameidx4input in range(
                    endpoints[0], min(endpoints[1], num_frames - 1)
                ):
                    lookahead_steps = min(
                        random.randint(1, max_pred_steps),
                        num_frames - 1 - frameidx4input,
                    )
                    frameidx4target = frameidx4input + lookahead_steps
                    fileidx4target = self.metadata[frameidx4target]["file_idx"]

                    filedesc4target = file_desc_cache.get(fileidx4target)
                    if filedesc4input is None and filedesc4target is None:
                        # both files need to be opened
                        file_descs = loop.run_until_complete(
                            asyncio.gather(
                                async_open_file(fileidx4input, None, self.recorddir),
                                async_open_file(
                                    fileidx4target, file_desc_cache, self.recorddir
                                ),
                            )
                        )
                        # order of return values are preserved. Ref: https://stackoverflow.com/questions/54668701/asyncio-gather-scheduling-order-guarantee
                        filedesc4input, filedesc4target = file_descs[0], file_descs[1]
                    elif filedesc4input is None:
                        # only need files for input frame
                        file_descs = loop.run_until_complete(
                            async_open_file(fileidx4input, None, self.recorddir)
                        )
                        filedesc4input = file_descs
                    elif filedesc4target is None:
                        # only need files for target frame
                        file_descs = loop.run_until_complete(
                            async_open_file(
                                fileidx4target, file_desc_cache, self.recorddir
                            )
                        )
                        filedesc4target = file_descs

                    start_time = perf_counter()
                    frame_pairs = loop.run_until_complete(
                        asyncio.gather(
                            async_read_frame(
                                filedesc4input,
                                self.metadata[frameidx4input],
                                input_features,
                            ),
                            async_read_frame(
                                filedesc4target,
                                self.metadata[frameidx4target],
                                target_features,
                            ),
                        )
                    )
                    end_time = perf_counter()
                    self.timer.add(
                        end_time - start_time,
                        frame_pairs[0].nbytes + frame_pairs[1].nbytes,
                    )
                    # print(self.timer.summarize())
                    yield {
                        "input": frame_pairs[0],
                        "target": frame_pairs[1],
                        "lookahead_steps": np.asarray(lookahead_steps),
                        "input_features": input_features,
                        "target_features": target_features,
                    }
                # close file descriptor for

                if filedesc4input is not None:
                    loop.run_until_complete(close_aiofile(filedesc4input))
        finally:
            # close open files
            loop.run_until_complete(close_files_in_cache(file_desc_cache))

            # wrap up async works
            loop.run_until_complete(loop.shutdown_asyncgens())
            loop.close()

    @staticmethod
    def prefetch_thread_worker(prefetch_data):
        # Lazily import to prevent circular import
        # shc: not sure what this is for?
        from torchdata.dataloader2 import communication

        itr = iter(prefetch_data.source_data_generator)
        stop_iteration = False
        while prefetch_data.run_prefetcher:
            if (
                len(prefetch_data.prefetch_buffer) < prefetch_data.buffer_size
                and not stop_iteration
            ):
                try:
                    item = next(itr)
                    prefetch_data.prefetch_buffer.append(item)
                except StopIteration:
                    stop_iteration = True
                # shc: probably not necessary for now
                except communication.iter.InvalidStateResetRequired:
                    stop_iteration = True
                except communication.iter.TerminateRequired:
                    prefetch_data.run_prefetcher = False
            elif stop_iteration and len(prefetch_data.prefetch_buffer) == 0:
                prefetch_data.run_prefetcher = False
            else:  # Buffer is full, waiting for main thread to consume items
                # TODO: Calculate sleep interval based on previous consumption speed
                time.sleep(PRODUCER_SLEEP_INTERVAL)

    def fetch_frame_pairs(
        self,
        input_features: List[str],
        target_features: List[str],
        max_pred_steps: int,
        prefetch_buffer_size: int = 10,
    ):

        if prefetch_buffer_size < 1:
            yield from self.iterate_frame_pairs(
                input_features, target_features, max_pred_steps
            )
        else:
            # ref: https://github.com/pytorch/data/blob/main/torchdata/datapipes/iter/util/prefetcher.py
            # preftech using a separate thread
            try:
                prefetch_data = _PrefetchData(
                    self.iterate_frame_pairs(
                        input_features, target_features, max_pred_steps
                    ),
                    prefetch_buffer_size,
                )
                thread = threading.Thread(
                    target=WSeqRecord.prefetch_thread_worker,
                    args=(prefetch_data,),
                    daemon=True,
                )
                thread.start()
                while prefetch_data.run_prefetcher:
                    if len(prefetch_data.prefetch_buffer) > 0:
                        yield prefetch_data.prefetch_buffer.popleft()
                    else:
                        # TODO: Calculate sleep interval based on previous availability speed
                        time.sleep(CONSUMER_SLEEP_INTERVAL)
            finally:
                prefetch_data.run_prefetcher = False
                if thread is not None:
                    thread.join()
                    thread = None

    def dump(self) -> None:
        """save attributes of instance of record into a pickled file and yaml file for visual inspection.

        Note:
        saving attribute dict instead of pickled class: pickling class and loading it is a mess because of
        path issues.
        """
        dic = copy.deepcopy(self.__dict__)
        # do not want to pickle a python module
        dic["file_desc"] = None
        with open(os.path.join(self.recorddir, "record.dict"), mode="wb") as f:
            pickle.dump(dic, file=f)

        # save some attributes of the seqrecord to yaml for human inspection
        dic["idx_range_of_files"] = None
        # transform some features to make them readable in yaml
        for key, val in dic["metadata"].items():
            for feature in dic["features_written"]:
                val[feature]["dtype"] = val[feature]["dtype"].str
                val[feature]["shape"] = list(val[feature]["shape"])
        with open(os.path.join(self.recorddir, "record_dict.yaml"), mode="w") as f:
            f.write("# Configs for human inspection only!\n")
            f.write(yaml.dump(dic))

    @classmethod
    def load_record_from_dict(cls, recorddir: str) -> WSR:
        """return an instance of sequence record from file that stores attributes of record as a
        dict (stored at path).

        Args:
            path (str): path to the file that stores dict of attributes of seqrecord

        Returns:
            WSR: an instance of record
        """

        file_path = os.path.join(recorddir, "record.dict")
        with open(file_path, mode="rb") as f:
            obj_dict = pickle.load(f)
        obj = cls(
            recorddir=recorddir,
        )
        obj_dict.pop("recorddir", None)
        for key, value in obj_dict.items():
            setattr(obj, key, value)
        return obj


async def async_open_file(file_idx: int, file_desc_cache: Optional[LRUCache], dir: str):
    file_desc = await aiofiles.open(recordfileidx2path(dir, file_idx), "rb")
    if file_desc_cache is not None:
        evicted = file_desc_cache.put(file_idx, file_desc)
        if evicted is not None:
            await evicted.close()
    return file_desc


async def close_files_in_cache(file_desc_cache: LRUCache) -> None:
    for key in file_desc_cache.keys():
        await file_desc_cache.pop(key).close()
    return None


async def close_aiofile(file_desc: io.BufferedReader) -> None:
    await file_desc.close()
    return


# notes: have to use file manager (instead of an lru function since we need to close files)
#        how to verify we are doing async?
# other approaches to be compared with:
#           1. read the whole frame and extract feature data (since reading small pieces of data multiple times is probably slow)
#           2. no async at all
async def async_read_frame(
    file_desc: io.BufferedReader, metadata_frame: dict, features: List[str]
) -> np.ndarray:
    """Given frame metadata and file object that contain frame data, read features from the frame data
    according to features

    Args:
        file_desc (io.BufferedReader): file object that contains the frame data (file object returned by aiofiles) is a subtype of BufferedReader
        metadata_frame (dict): _description_
        features (List[str]): _description_

    Returns:
        np.ndarray: _description_
    """
    await file_desc.seek(metadata_frame["bytes_offset"])
    data_bytes = await file_desc.read(metadata_frame["nbytes"])
    frame = {}
    # read the whole chunk or we read each file separately
    for feature in features:
        # b = file_desc.read(metadata[feature]["nbytes"])
        # array1d = np.frombuffer(
        #     bytes,
        #     dtype=metadata[feature]["dtype"],
        # )
        # frame[feature] = array1d
        # `await` halts `async_read_frame` and gives control back
        start = metadata_frame[feature]["bytes_offset"]
        end = start + metadata_frame[feature]["nbytes"]
        frame[feature] = np.frombuffer(
            data_bytes[start:end],
            dtype=metadata_frame[feature]["dtype"],
        ).reshape(metadata_frame[feature]["shape"])

    frame_array = np.vstack([frame[feature] for feature in features])
    return frame_array



"""A package for decoding and encoding each item in data file."""

import collections
import io
import os
import pickle
from typing import (
    Any,
    BinaryIO,
    Dict,
    Generator,
    List,
    Optional,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    Deque,
)
import numpy as np
import yaml
from seqrecord.utils import WriterBuffer
import copy
from collections import deque
import time
from seqrecord.utils import PRODUCER_SLEEP_INTERVAL, CONSUMER_SLEEP_INTERVAL
import threading

MAX_RECORDFILE_SIZE = 1e9  # 1e8, 100 mb, maximum size of a single record file

RSR = TypeVar("RSR", bound="RSeqRecord")


def recordfileidx2path(recorddir, recordfile_idx: int) -> str:
    return os.path.join(recorddir, f"records_{recordfile_idx}.bin")


# todo: add metadata for each episode


class RSeqRecord:
    """A serialization protocal that stores sequences of episodes into record files, while provides metadata of each episode/frame
    to read segments from dataset."""

    def __init__(
        self,
        recorddir: str,
    ) -> None:
        # folder that stores data in a separate directory (subfolder)
        self.recorddir: str = recorddir
        os.makedirs(self.recorddir, exist_ok=True)

        self.features_written = None
        self.num_bytes: int = 0  # number of bytes written into current record file
        self.file_idx: int = 0  # number of record file created for dataset
        # track the idx endpoints for each record file, [[start_idx, end_idx]], both are inclusive
        self.idx_range_of_files: List[Tuple[int, int]] = []
        # file object for current record file
        self.file_desc: Optional[BinaryIO] = None
        # serialization proto info of each data item
        self.metadata: Dict[int, dict] = {}
        # index of current data item to be processed
        self.frame_idx: int = 0

        # a cache dict that stores protocal info for each (segment_len, sub features)
        self.metadata_segment_cache: Dict[str, dict] = {}
        self.write_buffer = WriterBuffer()

    def get_recordfiles(self) -> List[str]:
        return [self.recordfile_idx_to_path(i) for i in range(self.file_idx)]

    def write_item(
        self,
        frame: Dict[str, np.ndarray],
        is_seq_start: bool,
    ) -> None:
        """write one item data dict(feature->np.ndarray) into bytes and write encoded bytes into
        current record files.

        Args:
            item (Dict[str, np.ndarray]): feature to data (np.ndarray)
            is_seq_start (bool): denote if the item is the beginning of a sequence
        """
        if self.features_written is None:
            self.features_written = [key for key in frame]
        if is_seq_start:
            self.seq_start()

        # get file name and start position for item
        self.metadata[self.frame_idx] = {
            "frame_idx": self.frame_idx,
            "file_idx": self.file_idx,
            "bytes_offset": self.num_bytes,
            "is_seq_start": is_seq_start,
        }
        num_bytes_in_frame = 0
        for feature, data in frame.items():
            self.metadata[self.frame_idx][feature] = {
                "is_none": (
                    data.dtype == np.dtype("O") and data == None
                ),  # this feature is essentially missing, and
                "dtype": data.dtype,
                "shape": data.shape,
                "bytes_offset": num_bytes_in_frame,
                "nbytes": data.nbytes,
            }
            self.write_buffer.write(data.tobytes())
            num_bytes_in_frame += data.nbytes

        self.num_bytes += num_bytes_in_frame
        self.frame_idx += 1

        return

    def seq_start(self) -> None:
        """Notify the record that a new sequence is being written, let the record decide if we need
        a new record file to write into.

        Two cases we need to open new file:
            1. we currently do not have record file to write into
            2. current file size is big enough (larger than MAX_RECORDFILE_SIZE)
        """
        if self.num_bytes > MAX_RECORDFILE_SIZE:
            # current record file big enough
            self.num_bytes = 0
            self.file_idx += 1
            self.idx_range_of_files[-1].append(self.frame_idx - 1)
            self.idx_range_of_files.append([self.frame_idx])
            self.file_desc.write(self.write_buffer.getbuffer())
            self.file_desc.flush()
            self.file_desc.close()
            self.write_buffer.clear()
            self.file_desc = open(
                recordfileidx2path(self.recorddir, self.file_idx),
                mode="wb",
            )
        elif self.file_desc == None:
            # no opened record file to write into
            self.idx_range_of_files.append([self.frame_idx])
            self.file_desc = open(
                recordfileidx2path(self.recorddir, self.file_idx), mode="wb"
            )

    def read_frame(
        self,
        file_desc: Union[io.BufferedReader, BinaryIO],
        metadata_frame: Dict[str, Union[int, dict]],
        features: List[str],
    ) -> Dict[str, np.ndarray]:
        """Given record file descriptor and serialization proto of a single data item, return the
        decoded dictionary(feature->data(np.ndarray)) of the item.

        Args:
            recordfile_desc (io.BufferedReader): python file object of the record file (required by numpy)
            itemproto (Dict[str, Any]): dict that contains protocal info of a specific data item

        Returns:
            Dict[str, np.ndarray]: data
        """
        frame = {}
        frame_offset = metadata_frame["bytes_offset"]
        for feature in features:
            frame[feature] = np.memmap(
                file_desc,
                dtype=metadata_frame[feature]["dtype"],
                mode="r",
                offset=frame_offset + metadata_frame[feature]["bytes_offset"],
                shape=metadata_frame[feature]["shape"],
            )
        # * do we need to close the memmap?
        return frame

    def read_frame_frombuffer(
        self,
        file_desc: Union[io.BufferedReader, BinaryIO],
        metadata_frame: Dict[str, Union[int, dict]],
        features: List[str],
    ) -> Dict[str, np.ndarray]:
        """Given record file descriptor and serialization proto of a single data item, return the
        decoded dictionary(feature->data(np.ndarray)) of the item, where decoding is done by
        np.frombuffer()

        Args:
            recordfile_desc (io.BufferedReader): python file object of the record file (required by numpy)
            itemproto (Dict[str, Any]): dict that contains protocal info of a specific data item

        Returns:
            Dict[str, np.ndarray]: data
        """
        frame = {}
        file_desc.seek(metadata_frame["bytes_offset"])
        for feature in features:
            bytes = file_desc.read(metadata_frame[feature]["nbytes"])
            array1d = np.frombuffer(
                bytes,
                dtype=metadata_frame[feature]["dtype"],
            )
            frame[feature] = array1d.reshape(metadata_frame[feature]["shape"])
        return frame

    def read_frames(
        self, features: List[str]
    ) -> Generator[Dict[str, np.ndarray], None, None]:
        """Given that the dataset has been recored, decode the record sequentially, each time
        returning a dict that contains the data item.

        Args:
            features [List[str]]: a list of features requested to read from item
            shuffle_recordfile: bool: if we shuffle at the record file level when reading items in the record
        Yields:
            Generator[Dict[str, np.ndarray], None, None]: data item [feature->data]. All data items are being returned sequentially
        """
        recordfiles = list(range(self.file_idx))
        for i in recordfiles:
            recordfile_path = recordfileidx2path(self.recorddir, i)
            endpoints = self.idx_range_of_files[i]
            with open(recordfile_path, mode="rb") as f:
                for idx in range(endpoints[0], endpoints[1] + 1):
                    item = self.read_frame(f, self.metadata[idx], features)
                    yield {feature: item[feature] for feature in features}

    def get_metadata4segment(
        self, segment_len: int, sub_features: Optional[List[str]] = None
    ) -> Dict[str, Any]:
        """Generate a protocal for reading segments from records. Each data item of segment should
        contain all features in sub_features.

        Note:
        Only call this function when record has scanned all data in dataset, and record has valid attributes: rootdir, recordfile_idx

        Args:
            segment_len (int): length of segment we are reading, 1< segment_len < sequence length
            sub_features: (Optional[List[str]]): features (modalities data) we need for each data item in segment to contain. If it is None,
            then we read all features.

        Returns:
            Dict[str, Any]: protocal needed for reading segments from data
        """

        def has_sub_features(itemproto: Dict[str, Any]) -> bool:
            return all(not itemproto[feature]["is_none"] for feature in sub_features)

        def update_segmentproto(item_idx: int, is_segment_start: bool) -> None:
            if is_segment_start:
                head4segment.append(item_idx)
            recordfile_idx = self.metadata[item_idx]["file_idx"]
            file2segment_items[recordfile_idx].append((is_segment_start, item_idx))
            return

        if sub_features is None:
            sub_features = list(self.features_written)
        else:
            assert all(
                feature in self.features_written for feature in sub_features
            ), "Unknow features requested"
        cache_key = str(segment_len) + "#" + "#".join(sorted(sub_features))
        if cache_key in self.metadata_segment_cache:
            return self.metadata_segment_cache[cache_key]
        head4segment: List[int] = []
        file2segment_items: dict[int, List[Tuple[bool, int]]] = collections.defaultdict(
            list
        )
        q = collections.deque()
        q_has_seg_tail = False  # indicates if the elements currently in queue are tail of some segment
        for idx in range(self.frame_idx):
            itemproto = self.metadata[idx]
            if (not has_sub_features(itemproto)) or (itemproto["is_seq_start"]):
                # new seq start
                while q:
                    if q_has_seg_tail:
                        update_segmentproto(q.popleft(), is_segment_start=False)
                    else:
                        q.popleft()
                q_has_seg_tail = False
                if has_sub_features(itemproto):
                    # a valid start of sequence
                    q.append(idx)
            else:
                q.append(idx)
                if len(q) == segment_len:
                    # claim: elements in the queue must be from the same sequence
                    update_segmentproto(q.popleft(), is_segment_start=True)
                    q_has_seg_tail = True

        if q and q_has_seg_tail:
            # front element in queue is need as last element of some segment
            update_segmentproto(q.popleft(), is_segment_start=False)

        # 1. new seq (including broken) added before queue pops out
        #       the remaining elements in queue are completely useless
        # 2. new seq (including broken) added after queue has popped out
        #       the remaining elements are not start of segment but are tails of some segment
        self.metadata_segment_cache[cache_key] = {
            "segment_len": segment_len,
            "features": sub_features,
            "head4segment": head4segment,
            "file2segment_items": file2segment_items,
        }
        return self.metadata_segment_cache[cache_key]

    def read_segments(self, segment_proto: dict):
        """Iterate through the whole records and return segments sequential.

        Yields:
            segment_proto: info on in given segment_len and features
        """
        segment_len = segment_proto["segment_len"]
        recordfile_ids = list(segment_proto["file2segment_items"].keys())
        for recordfile_idx in recordfile_ids:
            item_list = segment_proto["file2segment_items"][recordfile_idx]
            recordfile_path = recordfileidx2path(self.recorddir, recordfile_idx)
            q = collections.deque()
            with open(recordfile_path, mode="rb") as f:
                for is_segment_start, item_idx in item_list:
                    q.append(
                        (
                            is_segment_start,
                            self.read_frame(
                                f, self.metadata[item_idx], segment_proto["features"]
                            ),
                        )
                    )
                    while not q[0][0]:
                        q.popleft()
                    if len(q) == segment_len:
                        yield self.collate_items(q)
                        q.popleft()

    def read_one_segment(
        self,
        segment_len: int,
        head_idx: int,
    ) -> Dict[str, List[np.ndarray]]:
        """Read a segment (of lenght segment_len) starting from the item index being head_idx.

        Args:
            segment_len (int): length of segment we need to generate
            head_idx (int): item_idx of the head of the segment to be read.

        Returns:
            Dict[str, np.ndarray]: segment data
        """
        recordfile_path = recordfileidx2path(
            self.recorddir, self.metadata[head_idx]["recordfile_idx"]
        )
        q = []
        with open(recordfile_path, mode="rb") as f:
            for idx in range(head_idx, head_idx + segment_len):
                q.append(
                    (
                        idx == head_idx,
                        self.read_frame_frombuffer(f, self.metadata[idx]),
                    )
                )
        return self.collate_items(q)

    def collate_items(
        self, q: Sequence[Tuple[bool, dict]]
    ) -> Dict[str, List[np.ndarray]]:
        segment = {}
        features = q[0][1].keys()
        for feature in features:
            segment[feature] = [item[feature] for _, item in q]
        return segment

    def close_recordfile(self):
        """Close opened file descriptor!

        This needs to be called when finishes scanning over the dataset.
        """
        self.idx_range_of_files[-1].append(self.frame_idx - 1)
        self.file_desc.write(self.write_buffer.getbuffer())
        self.write_buffer.close()
        self.write_buffer = None
        self.file_desc.flush()
        self.file_desc.close()
        self.file_idx += 1
        self.file_desc = None

    def dump(self) -> None:
        """save attributes of instance of record into a file.

        Note:
        saving attribute dict instead of pickled class: pickling class and loading it is a mess because of
        path issues.
        """
        dic = copy.deepcopy(self.__dict__)
        with open(os.path.join(self.recorddir, "record.dict"), mode="wb") as f:
            pickle.dump(dic, file=f)

        # save some attributes of the seqrecord to yaml for human inspection
        dic["metadata_segment_cache"] = None
        for key, val in dic["metadata"].items():
            for feature in dic["features_written"]:
                val[feature]["dtype"] = val[feature]["dtype"].str
                val[feature]["shape"] = list(val[feature]["shape"])
        with open(os.path.join(self.recorddir, "record_dict.yaml"), mode="w") as f:
            f.write("# Configs for human inspection only!\n")
            f.write(yaml.dump(dic))

    @classmethod
    def load_record_from_dict(cls, recorddir: str) -> RSR:
        """return an instance of sequence record from file that stores attributes of record as a
        dict (stored at path).

        Args:
            path (str): path to the file that stores dict of attributes of seqrecord

        Returns:
            SR: an instance of record
        """

        file_path = os.path.join(recorddir, "record.dict")
        with open(file_path, mode="rb") as f:
            obj_dict = pickle.load(f)
        obj = cls(
            recorddir=recorddir,
        )
        obj_dict.pop("recorddir", None)
        for key, value in obj_dict.items():
            setattr(obj, key, value)
        return obj



NAME_TO_VAR = {
    "2m_temperature": "t2m",
    "10m_u_component_of_wind": "u10",
    "10m_v_component_of_wind": "v10",
    "mean_sea_level_pressure": "msl",
    "surface_pressure": "sp",
    "toa_incident_solar_radiation": "tisr",
    "total_precipitation": "tp",
    "land_sea_mask": "lsm",
    "orography": "orography",
    "lattitude": "lat2d",
    "geopotential": "z",
    "u_component_of_wind": "u",
    "v_component_of_wind": "v",
    "temperature": "t",
    "relative_humidity": "r",
    "specific_humidity": "q",
}

VAR_TO_NAME = {v: k for k, v in NAME_TO_VAR.items()}

SINGLE_LEVEL_VARS = [
    "2m_temperature",
    "10m_u_component_of_wind",
    "10m_v_component_of_wind",
    "mean_sea_level_pressure",
    "surface_pressure",
    "toa_incident_solar_radiation",
    "total_precipitation",
    "land_sea_mask",
    "orography",
    "lattitude",
]
PRESSURE_LEVEL_VARS = [
    "geopotential",
    "u_component_of_wind",
    "v_component_of_wind",
    "temperature",
    "relative_humidity",
    "specific_humidity",
]
ALL_LEVELS = [50, 250, 500, 600, 700, 850, 925, 1000]

NAME_LEVEL_TO_VAR_LEVEL = {}

for var in SINGLE_LEVEL_VARS:
    NAME_LEVEL_TO_VAR_LEVEL[var] = NAME_TO_VAR[var]

for var in PRESSURE_LEVEL_VARS:
    for l in ALL_LEVELS:
        NAME_LEVEL_TO_VAR_LEVEL[var + "_" + str(l)] = NAME_TO_VAR[var] + "_" + str(l)

VAR_LEVEL_TO_NAME_LEVEL = {v: k for k, v in NAME_LEVEL_TO_VAR_LEVEL.items()}

DEFAULT_PRESSURE_LEVELS = {
    "u": [1000, 850, 500],
    "v": [1000, 850, 500],
    "z": [1000, 850, 500, 50],
    "t": [850, 500],
    "r": [850, 500],
}


from collections import OrderedDict
import io
import os
from typing import Callable

PRODUCER_SLEEP_INTERVAL = 0.0001  # Interval between buffer fullfilment checks
CONSUMER_SLEEP_INTERVAL = (
    0.0001  # Interval between checking items availablitity in buffer
)


class LRUCache:
    """LRU cache using OrderedDict

    Returns:
        _type_: _description_
    """

    def __init__(self, capacity: int, monitoring: bool = False) -> None:
        self.capacity = capacity
        self.values = OrderedDict()  # Dict[int, io.BufferedReader] = {}
        self.monitoring = monitoring
        if self.monitoring:
            self.hits = 0
            self.misses = 0

    def put(self, key: int, value):
        """add (key, value) to cache following lru policy using ordered dict

        Args:
            ele (_type_): _description_

        Returns:
            evicted element in case there is post-processing needed.
        """
        evicted = None
        if key not in self.values:
            if len(self.values) == self.capacity:
                # popitem() returns key, value pair
                _, evicted = self.values.popitem(last=False)
        else:
            self.values.pop(key)
        self.values[key] = value
        return evicted

    def get(self, key: int):
        """retrieve element from cache

        Args:
            key (_type_): _description_

        Returns:
            _type_: _description_
        """
        res = None
        if key in self.values:
            self.values[key] = self.values.pop(key)
            res = self.values[key]
        if self.monitoring:
            self.hits = self.hits + (res is not None)
            self.misses = self.misses + (res is None)
        return res

    def pop(self, key: int):
        return self.values.pop(key)

    def keys(self):
        return [key for key in self.values.keys()]

    def stats(self) -> str:

        if self.monitoring:
            return f"Cache hitting rate is {self.hits / (self.hits + self.misses)} with overall number of reads {self.hits + self.misses}"
        else:
            return "Monitoring was not enabled for cache, hence no stats available"


class FileManager:
    """manager opening and closing of file descriptors, with a LRU Cache for opened files (remote files, connection)."""

    # todo: add stats for monitoring cache miss and hits

    def __init__(
        self,
        fileidx2path: Callable[[int], str],
        cache_capacity: int,
        monitoring: bool = False,
    ) -> None:
        self.fileidx2path = fileidx2path
        self.cache = LRUCache(cache_capacity, monitoring)

    def open_file(self, file_idx: int) -> io.BufferedReader:
        f = self.cache.get(file_idx)
        if f is None:
            # todo: make this a variable or function or something fixed
            file_path = self.fileidx2path(file_idx=file_idx)
            f = open(file_path, "rb")
            evicted = self.cache.put(key=file_idx, value=f)
            if evicted is not None:
                evicted.close()
        return f

    def close_all_files(self) -> None:
        """Close all open file descriptors in cache"""
        for key in self.cache.keys():
            self.cache.pop(key).close()
        return


class WriterBuffer:
    def __init__(self) -> None:
        self.buffer = io.BytesIO()

    def write(self, buffer) -> None:
        self.buffer.write(buffer)

    def is_empty(self) -> bool:
        return self.buffer.tell() == 0

    def getbuffer(self) -> memoryview:
        return self.buffer.getbuffer()

    def getvalue(self):
        return self.buffer.getvalue()

    def clear(self) -> None:
        self.buffer.flush()
        self.buffer.seek(0)

    def close(self) -> None:
        self.buffer.close()


class TimeTracker:
    def __init__(self) -> None:
        self.time: float = 0.0
        self.nbytes: int = 0

    def add(self, time: float, nbytes: int) -> None:
        self.time += time
        self.nbytes += nbytes

    def summarize(self) -> str:
        return f"Took {self.time} to read {self.nbytes} bytes, with average rate {self.nbytes / self.time} bytes/s"



"""Iterative datapipes toread weather dataset in seqrecord format"""

from typing import Callable, Dict, List, Optional, Tuple

import numpy as np
import torch
import torchdata.datapipes as dp
from tqdm import tqdm

from .constants import VAR_LEVEL_TO_NAME_LEVEL
from .seqrecord import WSeqRecord


@dp.functional_datapipe("weather_datapipe")
class WeatherDatapipeFromWSeqRecord(dp.iter.IterDataPipe):
    """A torch datapiple class that iteratively read frame pairs from weather dataset (encoded by WSeqRecord)."""

    def __init__(
        self,
        record: WSeqRecord,
        input_features: List[str],
        target_features: List[str],
        max_pred_steps: int,
    ) -> None:
        super().__init__()
        self.record = record
        self.input_features = input_features
        self.target_features = target_features
        self.max_pred_steps = max_pred_steps

    def __iter__(self):
        # naive iterate
        # yield from self.record.iterate_frame_pairs(
        #     self.input_features, self.target_features, self.max_pred_steps
        # )
        # prefetech
        # yield from self.record.fetch_frame_pairs(
        #     self.input_features,
        #     self.target_features,
        #     self.max_pred_steps,
        #     prefetch_buffer_size=10,
        # )
        yield from self.record.async_iterate_frame_pairs(
            self.input_features,
            self.target_features,
            self.max_pred_steps,
        )


def collate_fn(batch: List[Dict[str, np.ndarray]]) -> Dict[str, torch.Tensor]:
    collated_batch: Dict[str, torch.Tensor] = {}
    collated_batch["input"] = torch.from_numpy(
        np.stack([batch[i]["input"] for i in range(len(batch))], axis=0)
    ).unsqueeze(1)
    collated_batch["target"] = torch.from_numpy(
        np.stack([batch[i]["target"] for i in range(len(batch))], axis=0)
    )
    collated_batch["lookahead_steps"] = torch.from_numpy(
        np.stack([batch[i]["lookahead_steps"] for i in range(len(batch))], axis=0)
    ).to(collated_batch["input"].dtype)
    # todo: scale it: predict_ranges= hrs_each_step * predict_ranges / 100
    # we can put them in weather datapipe
    for feature in ["input_features", "target_features"]:
        collated_batch[feature] = [
            VAR_LEVEL_TO_NAME_LEVEL[v] for v in batch[0][feature]
        ]
    return collated_batch


def build_datapipes(
    datapipe: dp.iter.IterDataPipe,
    shuffle_buffer_size: Optional[int],
    batch_size: int,
    mappings: List[Callable],
) -> dp.iter.IterDataPipe:
    """Iteratively apply operations to datapipe: shuffle, sharding, map, batch, collator

    Args:
        datapipe (dp.datapipe.IterDataPipe): entry datapipe
        shuffle_buffer_size (Optional[int]): buffer size for pseudo-shuffle
        batch_size (int):
        mappings (List[Callable]): a list of transforms applied to datapipe, between sharding and batch

    Returns:
        dp.datapipe.IterDataPipe: transformed datapipe ready to be sent to dataloader
    """
    # Shuffle will happen as long as you do NOT set `shuffle=False` later in the DataLoader
    # https://pytorch.org/data/main/tutorial.html#working-with-dataloader
    if shuffle_buffer_size is not None:
        datapipe = datapipe.shuffle(buffer_size=shuffle_buffer_size)
    # sharding: Place ShardingFilter (datapipe.sharding_filter) as early as possible in the pipeline,
    # especially before expensive operations such as decoding, in order to avoid repeating these expensive operations across worker/distributed processes.
    datapipe = datapipe.sharding_filter()
    for i, mapping in enumerate(mappings):
        datapipe = datapipe.map(fn=mapping)
    # Note that if you choose to use Batcher while setting batch_size > 1 for DataLoader,
    # your samples will be batched more than once. You should choose one or the other.
    # https://pytorch.org/data/main/tutorial.html#working-with-dataloader
    datapipe = datapipe.batch(batch_size=batch_size, drop_last=True)
    datapipe = datapipe.collate(collate_fn=collate_fn)
    return datapipe



"""Iterative datapipes built from seqrecord"""


from typing import Callable, Dict, List, Optional

import numpy as np
import torch
import torchdata.datapipes as dp
from tqdm import tqdm

from .seqrecord import SeqRecord


@dp.functional_datapipe("video_datapipe")
class VideoDatapipeFromSeqRecord(dp.iter.IterDataPipe):
    """A torch datapiple class that iteratively read video(episode) segment from record files."""

    def __init__(
        self,
        record: SeqRecord,
        segment_len: int,
        features_rename: Dict[str, str],
        shuffle_recordfiles: bool = False,
    ) -> None:
        super().__init__()
        self.segmentproto = record.get_proto4segment(
            segment_len, [feature for feature in features_rename]
        )
        self.record = record
        self.features_rename = features_rename
        self.shuffle_recordfiles = shuffle_recordfiles

    def __iter__(self):
        for segment in self.record.read_segments(
            self.segmentproto, shuffle_recordfiles=self.shuffle_recordfiles
        ):
            res = {}
            for feature in self.features_rename:
                res[self.features_rename[feature]] = segment[feature]
            yield res


@dp.functional_datapipe("item_datapipe")
class ItemDatapipeFromSeqRecord(dp.iter.IterDataPipe):
    """A torch datapiple class that iteratively read item (frame) from record files."""

    def __init__(
        self,
        record: SeqRecord,
        features_rename: Dict[str, str],
        shuffle_recordfiles: bool = False,
    ) -> None:
        super().__init__()
        self.record = record
        self.features_rename = features_rename
        self.shuffle_recordfiles = shuffle_recordfiles

    def __iter__(self):
        res = {}
        for item in self.record.read_items(
            features=[feature for feature in self.features_rename],
            shuffle_recordfiles=self.shuffle_recordfiles,
        ):
            res = {}
            for feature in self.features_rename:
                res[self.features_rename[feature]] = item[feature]
            yield res


def collate_fn(batch: List[Dict[str, np.ndarray]]) -> Dict[str, torch.Tensor]:
    collated_batch: Dict[str, torch.Tensor] = {}
    for feature in batch[0]:
        collated_batch[feature] = torch.from_numpy(
            np.stack([batch[i][feature] for i in range(len(batch))], axis=0)
        )
    return collated_batch


def list2array(data_list: Dict[str, List[np.ndarray]]) -> Dict[str, np.ndarray]:
    """transform data from list of np.array to a single numpy array. Only needed for video datapipes.
    Args:
        data_np (Dict[str, List[np.ndarray]]): _description_
    Returns:
        Dict[str, np.ndarray]: _description_
    """
    data_array: Dict[str, np.ndarray] = {}
    for feature in data_list:
        data_array[feature] = np.stack(data_list[feature], axis=0)
    return data_array


def build_datapipes(
    datapipe: dp.iter.IterDataPipe,
    shuffle_buffer_size: Optional[int],
    batch_size: int,
    mappings: List[Callable],
) -> dp.iter.IterDataPipe:
    """Iteratively apply operations to datapipe: shuffle, sharding, map, batch, collator

    Args:
        datapipe (dp.datapipe.IterDataPipe): entry datapipe
        shuffle_buffer_size (Optional[int]): buffer size for pseudo-shuffle
        batch_size (int):
        mappings (List[Callable]): a list of transforms applied to datapipe, between sharding and batch

    Returns:
        dp.datapipe.IterDataPipe: transformed datapipe ready to be sent to dataloader
    """
    # Shuffle will happen as long as you do NOT set `shuffle=False` later in the DataLoader
    # https://pytorch.org/data/main/tutorial.html#working-with-dataloader
    if shuffle_buffer_size is not None:
        datapipe = datapipe.shuffle(buffer_size=shuffle_buffer_size)
    # sharding: Place ShardingFilter (datapipe.sharding_filter) as early as possible in the pipeline,
    # especially before expensive operations such as decoding, in order to avoid repeating these expensive operations across worker/distributed processes.
    datapipe = datapipe.sharding_filter()
    for i, mapping in enumerate(mappings):
        datapipe = datapipe.map(fn=mapping)
    # Note that if you choose to use Batcher while setting batch_size > 1 for DataLoader,
    # your samples will be batched more than once. You should choose one or the other.
    # https://pytorch.org/data/main/tutorial.html#working-with-dataloader
    datapipe = datapipe.batch(batch_size=batch_size, drop_last=True)
    datapipe = datapipe.collate(collate_fn=collate_fn)
    return datapipe



from setuptools import setup, find_packages

setup(
    name="seqrecord",
    version="0.1.4",
    description="fix more bugs in writing frames",
    author_email="shuhang0chen@gmail.com",
    maintainer_email="shuhang0chen@gmail.com",
    packages=find_packages(),
    install_requires=["numpy", "torch"],
)
